import { GridNode } from "@react-types/grid";
import { Key, MutableRefObject } from "react";
import { CollectionBase, Node, SelectionMode, Sortable, SortDescriptor } from "@react-types/shared";
import { GridState } from "@react-stately/grid";
import { TableCollection, TableHeaderProps, TableBodyProps, ColumnProps, RowProps, CellProps } from "@react-types/table";
import { MultipleSelectionStateProps } from "@react-stately/selection";
type mappedColumn<T> = GridNode<T> & {
    index: number;
    delta: number;
    calculatedWidth?: number;
};
export function getContentWidth(widths: Map<Key, number>): number;
export function isStatic(width: number | string): boolean;
export function parseStaticWidth(width: number | string, tableWidth: number): number;
export function getMaxWidth(maxWidth: number | string, tableWidth: number): number;
export function getMinWidth(minWidth: number | string, tableWidth: number): number;
export function getDynamicColumnWidths<T>(dynamicColumns: GridNode<T>[], availableSpace: number, tableWidth: number): mappedColumn<T>[];
export interface AffectedColumnWidth {
    /** The column key. */
    key: Key;
    /** The column width. */
    width: number;
}
export interface AffectedColumnWidths extends Array<AffectedColumnWidth> {
}
export interface ColumnResizeState<T> {
    /** A ref whose current value is the state of all the column widths. */
    columnWidths: MutableRefObject<Map<Key, number>>;
    /** Setter for the table width. */
    setTableWidth: (width: number) => void;
    /** Trigger a resize and recalc. */
    onColumnResize: (column: GridNode<T>, width: number) => void;
    /** Callback for when onColumnResize has started. */
    onColumnResizeStart: () => void;
    /** Callback for when onColumnResize has ended. */
    onColumnResizeEnd: () => void;
    /** Getter for column width. */
    getColumnWidth(key: Key): number;
    /** Getter for column min width. */
    getColumnMinWidth(key: Key): number;
    /** Getter for column max widths. */
    getColumnMaxWidth(key: Key): number;
    /** Boolean for if a column is being resized. */
    isResizingColumn: boolean;
}
export interface ColumnResizeStateProps<T> {
    /** Collection of existing columns. */
    columns: GridNode<T>[];
    /** Callback to determine what the default width of a column should be. */
    getDefaultWidth?: (props: any) => string | number;
    /** Callback that is invoked during the entirety of the resize event. */
    onColumnResize?: (affectedColumnWidths: AffectedColumnWidths) => void;
    /** Callback that is invoked when the resize event is ended. */
    onColumnResizeEnd?: (affectedColumnWidths: AffectedColumnWidths) => void;
    /** The default table width. */
    tableWidth?: number;
}
export function useTableColumnResizeState<T>(props: ColumnResizeStateProps<T>): ColumnResizeState<T>;
export interface TableState<T> extends GridState<T, TableCollection<T>> {
    /** A collection of rows and columns in the table. */
    collection: TableCollection<T>;
    /** Whether the row selection checkboxes should be displayed. */
    showSelectionCheckboxes: boolean;
    /** The current sorted column and direction. */
    sortDescriptor: SortDescriptor;
    /** Calls the provided onSortChange handler with the provided column key and sort direction. */
    sort(columnKey: Key, direction?: 'ascending' | 'descending'): void;
    /** A map of all the column widths by key. */
    columnWidths: MutableRefObject<Map<Key, number>>;
    /** Boolean for if a column is being resized. */
    isResizingColumn: boolean;
    /** Getter for column width. */
    getColumnWidth(key: Key): number;
    /** Getter for column min width. */
    getColumnMinWidth(key: Key): number;
    /** Getter for column max widths. */
    getColumnMaxWidth(key: Key): number;
    /** Trigger a resize and recalc. */
    onColumnResize: (column: GridNode<T>, width: number) => void;
    /** Runs at the start of resizing. */
    onColumnResizeStart: () => void;
    /** Triggers the onColumnResizeEnd prop. */
    onColumnResizeEnd: () => void;
    /** Need to be able to set the table width so that it can be used to calculate the column widths, this will trigger a recalc. */
    setTableWidth: (width: number) => void;
}
export interface CollectionBuilderContext<T> {
    showSelectionCheckboxes: boolean;
    selectionMode: SelectionMode;
    columns: Node<T>[];
}
export interface TableStateProps<T> extends CollectionBase<T>, MultipleSelectionStateProps, Sortable {
    /** Whether the row selection checkboxes should be displayed. */
    showSelectionCheckboxes?: boolean;
    /** Function for determining the default width of columns. */
    getDefaultWidth?: (props: any) => string | number;
    /** Callback that is invoked during the entirety of the resize event. */
    onColumnResize?: (affectedColumnWidths: AffectedColumnWidths) => void;
    /** Callback that is invoked when the resize event is ended. */
    onColumnResizeEnd?: (affectedColumnWidths: AffectedColumnWidths) => void;
}
/**
 * Provides state management for a table component. Handles building a collection
 * of columns and rows from props. In addition, it tracks row selection and manages sort order changes.
 */
export function useTableState<T extends object>(props: TableStateProps<T>): TableState<T>;
/**
 * A TableHeader is a container for the Column elements in a Table. Columns can be statically defined
 * as children, or generated dynamically using a function based on the data passed to the `columns` prop.
 */
export let TableHeader: <T>(props: TableHeaderProps<T>) => JSX.Element;
/**
 * A TableBody is a container for the Row elements of a Table. Rows can be statically defined
 * as children, or generated dynamically using a function based on the data passed to the `items` prop.
 */
export let TableBody: <T>(props: TableBodyProps<T>) => JSX.Element;
/**
 * A Column represents a field of each item within a Table. Columns may also contain nested
 * Column elements to represent column groups. Nested columns can be statically defined as
 * children, or dynamically generated using a function based on the `childColumns` prop.
 */
export let Column: <T>(props: ColumnProps<T>) => JSX.Element;
/**
 * A Row represents a single item in a Table and contains Cell elements for each column.
 * Cells can be statically defined as children, or generated dynamically using a function
 * based on the columns defined in the TableHeader.
 */
export let Row: (props: RowProps) => JSX.Element;
/**
 * A Cell represents the value of a single Column within a Table Row.
 */
export let Cell: (props: CellProps) => JSX.Element;
export { Section } from '@react-stately/collections';

//# sourceMappingURL=types.d.ts.map
